<!DOCTYPE html>
<html lang="{{ site.lang | default: "en-US" }}">
  <head>
    <meta charset='utf-8'>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <link rel="stylesheet" href="{{ '/assets/css/style.css?v=' | append: site.github.build_revision | relative_url }}">
    <link rel="stylesheet" type="text/css" href="{{ '/assets/css/print.css' | relative_url }}" media="print">
    <meta name="google-site-verification" content="y_Xk9Ji1Qbw9dcKOzLLxYUppKENI4vw00yqOy_WIUAs" />
    <!--[if lt IE 9]>
    <script src="//html5shiv.googlecode.com/svn/trunk/html5.js"></script>
    <![endif]-->

{% seo %}
    {% include head-custom.html %}
  </head>

  <body>
    <div id="container">
      <div class="inner">

        <header>
          <h1>{{ page.title | default: site.title | default: site.github.repository_name }}</h1>
          <h2>{{ page.description | default: site.description | default: site.github.project_tagline }}</h2>
        </header>
        <section id="downloads" class="clearfix">
          {% if site.show_downloads %}
          <a href="{{ site.github.zip_url }}" id="download-zip" class="button"><span>Download .zip</span></a>
          <a href="{{ site.github.tar_url }}" id="download-tar-gz" class="button"><span>Download .tar.gz</span></a>
          {% endif %}
	{% if site.github.public %}
	  {% if site.github.is_project_page %}
          <a href="{{ site.github.repository_url }}" id="view-on-github" class="button"><span>View on GitHub</span></a>
	  {% else %}
          <a href="{{ site.github.owner_url }}" id="view-on-github" class="button"><span>View on GitHub</span></a>
	  {% endif %}
	{% endif %}
        </section>
        <hr>
        <section id="main_content">
          {{ content }}
        </section>

        <footer>
        {% if site.github.is_project_page %}
          {{ site.title | default: site.github.repository_name }} is maintained by <a href="{{ site.github.owner_url }}">{{ site.github.owner_name }}</a><br>
        {% endif %}
          This page was generated by <a href="https://pages.github.com">GitHub Pages</a>.
        </footer>

      </div>
    </div>

    <script type="module">
      Array.from(document.getElementsByClassName("language-mermaid")).forEach(element => {
        element.classList.add("mermaid");
      });
      import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@11/dist/mermaid.esm.min.mjs';
      mermaid.initialize({ startOnLoad: true, theme: "dark" });
    </script>

  </body>
</html>